require 'optparse'
require 'yaml'

module Postfix::Rate::Policyd

  DEFAULT_CORE_CONFIGURATION = {
    'config_file'  =>  '/etc/postfix-rate/policyd.yml',
    'trackers_file' =>  '/etc/postfix-rate/trackers.yml',
    'database_url' => nil,
    'daemonize'    => true,
    'bind_address' => '::',
    'bind_port'    => 10031,
    'log'          => 'stdout',
    'log_destination'  => '/var/log/postfix-rate/policyd.log',
    'pid_file'     => '/var/run/postfix-rate/policyd.pid',
    'log_level'    => :info,
    'debug'        => false,
    'command'      => 'start',
    'chdir'        => Dir.getwd,
  }.freeze

  DEFAULT_COMMANDS = [ 'start', 'setup' ].freeze

  class Configuration

    attr_reader :data

    def initialize(argv)

      config_file = DEFAULT_CORE_CONFIGURATION['config_file']
      trackers_file = DEFAULT_CORE_CONFIGURATION['trackers_file']
      options = {}
      @data = DEFAULT_CORE_CONFIGURATION.dup


      opts = OptionParser.new do |opts|
        opts.banner = usage

        opts.on("-c", "--configuration FILE", "Sets the configuraton file") do |file|
          options['config_file'] = config_file = file
        end

        opts.on("-t", "--trackers FILE", "Initialize trackers from file") do |file|
          options['trackers_file'] = trackers_file = file
        end

        opts.on("-l", "--log-level LEVEL", "Specify the log level (fatal, error, warn, info, debug).") do |level|
          options['log_level'] = level.to_sym
        end

        opts.on("--log TRANSPORT", "Specify the log transport (stdout, file, syslog).") do |transport|
          options['log'] = transport
        end

        opts.on("--log-destination FILE|URL", "Specify the log file path or the syslog URL") do |file|
          options['log_file'] = file
        end

        opts.on("--pid-file FILE", "Specify the pid file path") do |file|
          options['pid_file'] = file
        end

        opts.on("--pid-file FILE", "Specify the pid file path") do |file|
          options['pid_file'] = file
        end

        opts.on("--no-daemonize", "Do not daemonize") do
          options['daemonize'] = false
        end

        opts.on("--debug", "Activate debugging mode") do
          options['debug'] = true
          options['daemonize'] = false
          options['log_level'] = :debug
          options['log'] = 'stdout'
        end
      end

      opts.parse! argv
      puts "= reading configuration from '#{config_file}'."

      raw_config = File.exists?(config_file) ? (YAML.load_file(config_file) || {}) : {}
      trackers = File.exists?(trackers_file) ? (YAML.load_file(trackers_file) || {}) : nil

      data.merge!(raw_config).merge!(options)
      if trackers
        data.merge!("trackers" => trackers)
      end

      data['command'] = argv.shift if argv
      data['argument'] = argv

      check_command
      check_data

      #ap data
    end

    def check_command
      if !command
        puts usage
        exit
      end

      if !DEFAULT_COMMANDS.include?(command)
        puts "\nUnknown command '#{command}'"
        puts usage
        exit
      end
    end

    def check_data
      raise ArgumentError, "Database URL is required." unless database_url
    end

    def host
      data['bind_address']
    end

    def port
      data['bind_port']
    end

    def database_url
      data['database_url']
    end

    def daemonize?
      data['daemonize']
    end

    def arguments
      data['arguments']
    end

    def command
      data['command']
    end

    def log
      data['log']
    end

    def log_destination
      data['log_destination']
    end

    def log_level
      data['log_level']
    end

    def debug?
      data['debug']
    end

    def policies
      data['policies'] || {}
    end

    def trackers
      data['trackers'] || {}
    end

    def usage
      u =  '-'*80 + "\n"
      u << "= postfix-rate-policyd: a postfix filter implementing rate policies\n\n"
      u << "usage:\n"
      u << "postfix-rate-policyd [params] (start|setup)\n"
      u << '-'*80 + "\n"

    end

  end

end