require 'celluloid/io'
require 'celluloid/autostart'

module Postfix::Rate::Policyd

  class Request
    attr_accessor :instance
    attr_accessor :sender
    attr_accessor :recipient
    attr_accessor :authenticated_identity
    attr_accessor :client_address
    attr_accessor :helo

    attr_accessor :recipient_count
    attr_accessor :size

    attr_accessor :entity
    attr_accessor :entity_address

    attr_reader   :attributes

    def initialize
      @attributes = {}
    end

    def set_attribute(attribute, value)
      @attributes[attribute] = value
    end

    def new_instance?
      instance != attributes['instance']
    end

    def new_instance!
      @instance =                attributes.fetch('instance')
      @sender =                  attributes.fetch('sender', '-unknown-')
      @recipient  =              attributes.fetch('recipient', '-unknown-')
      @authenticated_identity =  attributes.fetch('sasl_username', nil)
      @client_address =          attributes.fetch('client_address')
      @size =                    attributes.fetch('size', 0).to_i
      @helo =                    attributes.fetch('helo_name', '-unknown-')
      @recipient_count =         attributes.fetch('recipient_count', 0).to_i
    end

    def authenticated?
      !@authenticated_identity.blank?
    end

    def protocol_state
      @state = attributes.fetch('protocol_state', nil)
    end

    def check_state?(requested_state)
      protocol_state == requested_state
    end

    def additional_info
      info = []
      info << "helo: '#{helo}'"
      info << "identity: '#{authenticated_identity}'" if authenticated?
      info << "entity: '#{entity.label}" unless entity.blank?

      info.join(', ')
    end

    def label
      "[#{client_address}]: #{sender} -> #{recipient} (#{additional_info})"
    end

  end


  class RequestWorker
    include Celluloid
    include Celluloid::Internals::Logger
    def debug(string)
      Celluloid.logger.debug(string) if Celluloid.logger
    end    

    finalizer :shut_worker_down

    attr_reader :client
    attr_reader :request
    attr_accessor :backend

    attr_reader :host
    attr_reader :port

    #
    # initializes a new worker
    #
    def initialize
    end

    def initialize_data_backend
      @backend ||= Postfix::Rate::Policyd::DataBackend.new
    end
    
    #
    #
    #
    def process_request(client)
      initialize_data_backend
      new_client(client)
      loop { break if handle_line(client.readline.chomp!) }
    rescue EOFError
      debug "= #{host}:#{port} disconnected"
      close_connection
    end

    def send_response(data)
      client.write data
    end

    def close_connection
      debug "= #{host}:#{port} closing"
      client.close unless client.closed?
    end



    def shut_worker_down
      debug "= #{host}:#{port} shutting down"
    end

    private


    #
    # A new client has connected
    #
    def new_client(client)
      @client = client
      @request = Request.new
      _, @port, @host = client.peeraddr
      debug "= Received connection from #{host}:#{port}"
    end


    #
    # for each complete request
    #
    def handle_request
      debug("---- new request----")
      debug request.attributes.inspect

      if request.new_instance?

        start_session
        res = check_initial_quota
        return res if res
      end

      case request.protocol_state
      when 'RCPT'
        res = check_rcpt_quota
        return res if res
      when 'END-OF-MESSAGE'
        res = check_final_quota_and_update
        return res if res
      end

      'OK'
    end

    #
    #
    #
    def start_session
      request.new_instance!
      backend.fetch_entity(request)
    end


    def check_initial_quota
      response = backend.check_initial_quota(request)
      if response.nack?
        return "DEFER #{response.message}"
      end
      nil
    end

    def check_rcpt_quota
      response = backend.check_rcpt_quota(request)
      if response.nack?
        return "DEFER #{response.message}"
      end
      nil
    end

    def check_final_quota_and_update
      response = backend.check_final_quota_and_update(request)
      if response.nack?
        return "DEFER #{response.message}"
      end
      nil
    end

    #
    #
    #
    def handle_line(text)
      #debug "#{host}:#{port} > #{text}"
      if text == ''
        resp = handle_request

        send_response("action=#{resp}\n\n")
        @request = Request.new
      else
        match = /^(.*?)=(.*)$/.match(text)

        if !match
          warn("- Protocol error '#{text}'")
          close_connection
          return true
        end

        request.set_attribute match[1],  match[2]
      end
      false
    rescue Exception => e
      Celluloid::Internals::Logger.crash '- An error occurred', e
      close_connection
      true
    end

  end
end
